package test.migrate;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import schedframe.events.scheduling.SchedulingEvent;
import schedframe.resources.ResourceStatus;
import schedframe.resources.computing.ComputingResource;
import schedframe.resources.computing.Node;
import schedframe.resources.computing.Processor;
import schedframe.resources.units.ProcessingElements;
import schedframe.resources.units.ResourceUnit;
import schedframe.resources.units.ResourceUnitName;
import schedframe.resources.units.StandardResourceUnitName;
import schedframe.scheduling.manager.resources.ClusterResourceManager;
import schedframe.scheduling.manager.resources.ResourceManager;
import schedframe.scheduling.manager.tasks.JobRegistry;
import schedframe.scheduling.manager.tasks.JobRegistryImpl;
import schedframe.scheduling.plan.SchedulingPlanInterface;
import schedframe.scheduling.plan.impl.SchedulingPlan;
import schedframe.scheduling.plugin.ModuleList;
import schedframe.scheduling.queue.TaskQueue;
import schedframe.scheduling.queue.TaskQueueList;
import schedframe.scheduling.tasks.TaskInterface;
import dcworms.schedframe.scheduling.ExecTask;
import example.localplugin.BaseLocalSchedulingPlugin;
import gridsim.dcworms.DCWormsTags;

public class FCFSBF_RandomPluginMigrate extends BaseLocalSchedulingPlugin {

	public SchedulingPlanInterface<?> schedule(SchedulingEvent event, TaskQueueList queues, JobRegistry jobRegistry,
			ResourceManager resManager, ModuleList modules) {

		ClusterResourceManager resourceManager = (ClusterResourceManager) resManager;
		List<Processor> processors = resourceManager.getProcessors();
		SchedulingPlan plan = new SchedulingPlan();
		// choose the events types to serve.
		// Different actions for different events are possible.
		switch (event.getType()) {
		case START_TASK_EXECUTION:
		case TASK_FINISHED:
		//case TIMER:
			// our tasks are placed only in first queue (see BaseLocalSchedulingPlugin.placeJobsInQueues() method)
			TaskQueue q = queues.get(0);
			// check all tasks in queue

			for (int i = 0; i < q.size(); i++) {
				TaskInterface<?> task = q.get(i);
				// if status of the tasks in READY
				if (task.getStatus() == DCWormsTags.READY) {
					Map<ResourceUnitName, ResourceUnit> choosenResources = chooseResourcesForExecution(processors, task);
					if (choosenResources != null) {
						addToSchedulingPlan(plan, task, choosenResources);
					} 
				}
			}
			break;
		case POWER_LIMIT_EXCEEDED:
			String src = event.getSource();
			optimizeEnergyUsage(jobRegistry, resourceManager , src);
			break;

		}
		
		return plan;
	}

	private void optimizeEnergyUsage(JobRegistry jobRegistry, ClusterResourceManager resourceManager, String resName) {
		Node overLoadedNode = (Node)resourceManager .getResourceByName(resName);
		Node leastLoadedNode = findLeastLoadedResource(resourceManager.getNodes());
		
		if(leastLoadedNode.getLoadInterface().getRecentUtilization().getValue() >= overLoadedNode.getLoadInterface().getRecentUtilization().getValue()){
			return;
		}

		JobRegistry srcHostJr = new JobRegistryImpl(overLoadedNode.getFullName());
		int runningTasksNumber = srcHostJr.getRunningTasks().size();
		for(int i = srcHostJr.getRunningTasks().size() - 1; i >= 0 && runningTasksNumber > 2; i--) {
			ExecTask execTask = srcHostJr.getRunningTasks().get(i);
			Map<ResourceUnitName, ResourceUnit> destinationResources = chooseResourcesForExecution(findLeastLoadedResource(resourceManager.getNodes()).getProcessors(), execTask);
			jobRegistry.migrateTask(execTask.getJobId(), execTask.getId(), destinationResources);
			runningTasksNumber--;
		}
	}

	private Map<ResourceUnitName, ResourceUnit> chooseResourcesForExecution(
			List<Processor> processors, TaskInterface<?> task) {

		Map<ResourceUnitName, ResourceUnit> map = new HashMap<ResourceUnitName, ResourceUnit>(1);

		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 0;
		}

		if (cpuRequest != 0) {

			if (processors.size() < cpuRequest) {
				return null;
			}

			List<ComputingResource> choosenResources = new ArrayList<ComputingResource>(cpuRequest);				
			for (int i = 0; i < processors.size() && cpuRequest > 0; i++) {
				if (processors.get(i).getStatus() == ResourceStatus.FREE) {
					choosenResources.add(processors.get(i));
					cpuRequest--;
				}
			}
			if (cpuRequest > 0) {
				return null;
			}

			ProcessingElements pe = new ProcessingElements();
			pe.addAll(choosenResources);
			map.put(StandardResourceUnitName.PE, pe);
			return map;
		}

		return null;
	}

	private Node findLeastLoadedResource(List<Node> nodes ){
		Node leastLoadedNode = null;
		double minLoad = Double.MAX_VALUE;
		
		for(int i = 0; i < nodes.size(); i++){
			Node node = nodes.get(i);
			double totalLoad = node.getLoadInterface().getRecentUtilization().getValue();
			if(totalLoad < minLoad){
				leastLoadedNode= node;
				minLoad = totalLoad;
			}
		}
		return leastLoadedNode;
	}

}
